"""
Logging Related Utils
"""

import os
import logging
from logging import (
    DEBUG,
    INFO,
    WARNING,
)
import warnings

log_levels = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
log_level_strs = log_levels.keys()

default_logging_level = "warning"
_default_handler = None


def _get_lib_name() -> str:
    """Get Library Name"""
    return __name__.split(".", maxsplit=1)[0]


def _get_lib_default_logging_level() -> int:
    """
    Get Default logging level, overridden by env variable `LLMSEARCH_VERBOSITY`
    the env var can take one of the following values - ["debug", "info", "warning", "error", "critical"]
    """
    env_var = "LLMSEARCH_VERBOSITY"
    env_value = os.getenv(env_var, None)
    if env_value is not None:
        if env_value not in log_level_strs:
            warnings.warn(
                f"Env variable - {env_var} should be one of - {list(log_levels.keys())}"
            )
            return log_levels[default_logging_level]
        return log_levels[env_value]
    return log_levels[default_logging_level]


def _get_library_root_logger() -> logging.Logger:
    """Retrives library root logger"""
    return logging.getLogger(_get_lib_name())


def _configure_library_root_logger() -> logging.Logger:
    """Configure library root logger"""
    global _default_handler  # pylint: disable=global-statement

    if _default_handler:
        # Already configured library root logger
        return
    formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d - %(name)s:%(lineno)d - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    _default_handler = logging.StreamHandler()
    _default_handler.setFormatter(formatter)
    library_root_logger = logging.getLogger(_get_lib_name())
    library_root_logger.addHandler(_default_handler)
    library_root_logger.setLevel(_get_lib_default_logging_level())
    # when running the api through scikit-learn (for hyperparameter search), logs get propogated to some root logger
    # so disabling by default
    library_root_logger.propagate = False


def get_logger(name: str) -> logging.Logger:
    """Gets a logger if name is supplied or returns the library root logger"""
    if name:
        return logging.getLogger(name)
    return _get_library_root_logger()


def set_verbosity(verbosity: int):
    """Sets the logging verbosity of the library root logger"""
    _get_library_root_logger().setLevel(verbosity)


def set_verbosity_info():
    """Sets the logging verbosity of the library root logger to info"""
    return set_verbosity(INFO)


def set_verbosity_warning():
    """Sets the logging verbosity of the library root logger to warning"""
    return set_verbosity(WARNING)


def set_verbosity_debug():
    """Sets the logging verbosity of the library root logger to debug"""
    return set_verbosity(DEBUG)


def disable_default_handler():
    """Disable default handler"""
    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().removeHandler(_default_handler)


def enable_default_handler():
    """Enable default handler"""
    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().addHandler(_default_handler)


def add_handler(handler: logging.Handler):
    """Add a custom handler to the library root logger"""
    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().addHandler(handler)


def remove_handler(handler: logging.Handler):
    """Remove a handler to the library root logger"""
    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().removeHandler(handler)


# setup library root logger
_configure_library_root_logger()



# pylint: skip-file
"""
Common Utilties for Models
"""
import gc
import math
import random
from itertools import islice
from typing import List, Dict, Union, Tuple, Iterable, Iterator

import torch
import numpy as np
from tqdm.auto import tqdm
from transformers import AutoTokenizer, AutoModel

from llmsearch.utils.gen_utils import identify_and_validate_gen_params
from llmsearch.utils.logging_utils import get_logger
from llmsearch.utils.mem_utils import batch_without_oom_error, gc_cuda

logger = get_logger(__name__)


def get_device() -> str:
    """Get device one of "cpu", "cuda", "mps"

    Returns:
        str: device str
    """
    if torch.backends.mps.is_built() and torch.backends.mps.is_available():
        return "mps"
    if torch.cuda.is_available():
        return "cuda"
    return "cpu"


def seed_everything(seed: int):
    """Seed for reproducibilty"""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


@batch_without_oom_error
def run_inference(
    model: AutoModel,
    tokenizer: AutoTokenizer,
    is_encoder_decoder: bool,
    batch_size: int,
    disable_batch_size_cache: bool,  # pylint: disable=unused-argument
    device: str,
    model_inputs: List,
    tokenizer_encode_args: Dict,
    tokenizer_decode_args: Dict,
    generation_args: Dict = None,
    disable_generation_param_checks: bool = False,
    return_optimal_batch_size: bool = False,
    output_preproc: callable = lambda x: x.strip(),
    callbacks: List = None,
) -> Union[Tuple[List, int], List]:
    """Infer on data with a specific batch size

    Args:
        model (AutoModelForSeq2SeqLM): model with a `.generate` method
        tokenizer (AutoTokenizer): tokenizer to tokenize the input
        is_encoder_decoder (bool): whether the model is an encoder-decoder model, `False` if not
        batch_size (int): batch_size to run inference with, this gets dynamically reduced if the inference function encounters OOM errors
        disable_batch_size_cache (bool): If `True` for each cross validation run, the pre-defined `batch_size` is used, this could lead to wasted computation time if OOM is raised
        device (str): device to run the inference on
        model_inputs (List): model inputs to do inference on
        tokenizer_encode_args (Dict): Encoding arguments for the `tokenizer`
        tokenizer_decode_args (Dict, optional): Decoding arguments for the `tokenizer`. Defaults to `{'skip_special_tokens' : True}`.
        generation_args (Dict, optional): generation kwargs to use while generating the output. Defaults to `None`.
        disable_generation_param_checks (bool, optional): Disables the custom generation parameter checks, this check does a sanity check of the parameters & produces warnings before doing generation, Not stable right now.
        return_optimal_batch_size (bool, optional): if the function should return the optimal batch size found, useful for caching when performing cross validation. Defaults to `False`.
        output_preproc (Callable, optional): Prepoc to run on the completion, by default strips the output. Note that this is applied on the completion. Defaults to `False`.
        callbacks (List, optional): List of callbacks to run after each generation, by default `None`.

    Returns:
        Union[Tuple[List, int], List]: outputs and or best batch size
    """
    seed = None

    tokenizer_decode_args = (
        tokenizer_decode_args
        if tokenizer_decode_args
        else {"skip_special_tokens": True}
    )

    outputs = []

    if not disable_generation_param_checks:
        generation_type = identify_and_validate_gen_params(gen_params=generation_args)
        logger.info("Detected generation type - %s", generation_type)

    if "generation_seed" in generation_args:
        seed = generation_args["generation_seed"]
        seed_everything(seed=seed)

    for _, batch in tqdm(
        enumerate(batcher(iterable=model_inputs, batch_size=batch_size)),
        total=math.ceil(len(model_inputs) / batch_size),
    ):
        gc.collect()
        gc_cuda()
        encoded_input = tokenizer(
            text=batch, **tokenizer_encode_args, return_tensors="pt"
        )
        # Useful to find completion
        decoded_input = tokenizer.batch_decode(
            encoded_input["input_ids"],
            **tokenizer_decode_args,
        )

        input_ids = encoded_input.input_ids.to(device)
        attention_mask = encoded_input.attention_mask.to(device)

        output_ids = model.generate(
            inputs=input_ids, attention_mask=attention_mask, **generation_args
        )

        if callbacks:
            for callback in callbacks:
                callback()

        decoded_output = tokenizer.batch_decode(
            sequences=output_ids,
            **tokenizer_decode_args,
        )

        # remove prompt
        if not is_encoder_decoder:
            decoded_output = decoder_parser(
                outputs=decoded_output,
                formatted_prompts=decoded_input,
                prepoc=output_preproc,
            )
        else:
            decoded_output = encoder_decoder_parser(
                outputs=decoded_output, prepoc=output_preproc
            )
        outputs.extend(decoded_output)

    for x, y_pred in zip(model_inputs, outputs):
        logger.debug("Input - %s", repr(x))
        logger.debug("Model Output - %s", repr(y_pred))

    if return_optimal_batch_size:
        return outputs, batch_size

    return outputs


def batcher(iterable: Iterable, batch_size: int) -> Iterator:
    """Batch a iterable into batches of `batch_size`

    Args:
        iterable (Iterable): iterable
        batch_size (int): batch size

    Yields:
        Iterator: iterator over batches
    """
    iterator = iter(iterable)
    while batch := list(islice(iterator, batch_size)):
        yield batch


def encoder_decoder_parser(outputs: str, prepoc: callable):
    """Applies the prepoc function on the completion

    Args:
        outputs (str): model outputs
        prepoc (callable): prepoc function

    Returns:
        List: processed outputs
    """
    return [prepoc(output) for output in outputs]


def decoder_parser(outputs: List[str], formatted_prompts: List[str], prepoc: callable):
    """Removes the prompt from the text and calls `prepoc` on the completion

    Args:
        outputs (List[str]): model outputs
        formatted_prompts (List[str]): formatted prompts
        prepoc (callable): prepoc function

    Returns:
        List: processed outputs
    """
    ret_val = []
    for output, formatted_prompt in zip(outputs, formatted_prompts):
        ret_val.append(prepoc(output[len(formatted_prompt) :]))
    return ret_val



"""
Inspired from [toma](https://github.com/BlackHC/toma), Memory related utils to do friendly inference.
"""

import gc
import math
import time
import ctypes
import inspect
import traceback

from functools import wraps
from typing import Any, Union, Tuple

import torch
import psutil

from llmsearch.utils.logging_utils import get_logger

logger = get_logger(__name__)


try:
    import pynvml
except ImportError:
    if torch.cuda.is_available():
        logger.warning(
            "pynvml package not found in a CUDA environment, Install it by running - `pip install nvidia-ml-py3` for setting caching of batch size while running hyp param search"
        )


class Cache:
    """Cache to store the optimal batch size for a specific configuration call using traceback and memory information"""

    def __init__(self):
        """Initializes cache"""
        self.cache = {}

    def set_value(
        self,
        value: Any,
        stacktrace: Tuple,
        total_available_gpu_memory: float,
        total_available_ram_memory: float,
    ):
        """Sets a value based on certain combination of different hashes

        Args:
            value (Any): hash value
            stacktrace (Tuple): stacktrace of the method call
            total_available_gpu_memory (float): total available gpu memory
            total_available_ram_memory (float): total available ram memory
        """
        # Create hash key
        hash_key = (stacktrace, total_available_gpu_memory, total_available_ram_memory)
        # Set value
        self.cache[hash_key] = value

    def get_value(
        self,
        current_value: Any,
        stacktrace: Tuple,
        total_available_gpu_memory: float,
        total_available_ram_memory: float,
    ) -> float:
        """Tries to get a value for a particular set of hashes, returns `current_value` (this only happens during the initial call)

        Args:
            current_value (Any): current value
            stacktrace (Tuple): stacktrace of the method call
            total_available_gpu_memory (float): total available gpu memory
            total_available_ram_memory (float): total available ram memory

        Returns:
            float: `current_value` if hash_key is not present, else returns the hashed key
        """
        hash_key = (stacktrace, total_available_gpu_memory, total_available_ram_memory)
        # Checks if current configuration in cache
        if hash_key in self.cache:
            val = self.cache[hash_key]
            return val
        return current_value

    def is_empty(self) -> bool:
        """Checks if the cache is empty

        Returns:
            bool: empty or not
        """
        return not bool(len(self.cache))

    def empty_cache(self):
        """Empties cache"""
        self.cache = {}


def get_traceback(ignore_first: int = 0, stack_context: int = 5) -> Tuple[Tuple]:
    """Get traceback from first to latest call

    Args:
        ignore_first (int, optional): ignore first n traceback. Defaults to 0.
        stack_context (int, optional): context for traceback. Defaults to 5.

    Returns:
        Tuple[Tuple]: Tuples of Function call and code
    """
    stack = inspect.stack(context=1)[ignore_first : ignore_first + stack_context]
    # Tuple of function and calling code
    simple_traceback = tuple((fi.function, fi.code_context[0]) for fi in stack)
    return simple_traceback


# init cache
cache = Cache()


def batch_without_oom_error(func: callable):
    """Perform Inference on a batch of samples by dividing the batch_size by 2 each time whenever OOM error happens, Function should have a `batch_size` and `disable_batch_size_cache` parameter

    Args:
        func (Callable): function having this signature of arguments `*args, batch_size, disable_batch_size_cache, **kwargs`
    """

    @wraps(func)
    def inner_wrapper(*args, batch_size: int, disable_batch_size_cache: bool, **kwargs):
        """When cache is turned on (`disable_batch_size_cache` - True), a cached batch_size can be used based on the configuration

        Args:
            batch_size (int): initial batch size
            disable_batch_size_cache (bool): setting this to `True` forces the inference to happen using `batch_size`

        """

        # if batch size cache is to be disabled
        if disable_batch_size_cache:
            # Empty cache
            if not cache.is_empty():
                cache.empty_cache()
        else:
            if not cache.is_empty():
                stacktrace = get_traceback(ignore_first=20, stack_context=10)
                gpu_info = get_gpu_information()
                total_available_gpu_memory = gpu_info[2] if gpu_info else gpu_info
                total_available_ram_memory = get_total_available_ram()
                # Get cached batch size if present
                batch_size = cache.get_value(
                    current_value=batch_size,
                    stacktrace=stacktrace,
                    total_available_gpu_memory=total_available_gpu_memory,
                    total_available_ram_memory=total_available_ram_memory,
                )
        logger.info(
            "Starting inference with generation parameters - %s",
            kwargs.get("generation_args", {}),
        )
        logger.info("Performing inference with batch_size - %s", batch_size)
        start_time = time.time()
        while True:
            # Try running with specified batch size
            try:
                res = func(
                    *args,
                    batch_size=batch_size,
                    disable_batch_size_cache=disable_batch_size_cache,
                    **kwargs,
                )
                gc_cuda()
                if not disable_batch_size_cache:
                    stacktrace = get_traceback(ignore_first=20, stack_context=10)
                    gpu_info = get_gpu_information()
                    total_available_gpu_memory = gpu_info[2] if gpu_info else gpu_info
                    total_available_ram_memory = get_total_available_ram()
                    # Set value for next iteration with the input hash
                    logger.debug(
                        "Setting batch_size cache value - %s for this particular configuration",
                        batch_size,
                    )
                    cache.set_value(
                        value=batch_size,
                        stacktrace=stacktrace,
                        total_available_gpu_memory=total_available_gpu_memory,
                        total_available_ram_memory=total_available_ram_memory,
                    )
                end_time = time.time()
                latency = end_time - start_time
                logger.info("Finished running inference, took %f secs", latency)
                return res
            except RuntimeError as exception:
                if batch_size > 1 and should_reduce_batch_size(exception):
                    logger.info(
                        "Unable to fit batch size - %d, Reducing batch size to - %d",
                        batch_size,
                        batch_size // 2,
                    )
                    batch_size //= 2
                    gc_cuda()
                elif isinstance(exception, NotImplementedError):
                    raise exception
                else:
                    raise Exception(  # pylint: disable=broad-exception-raised
                        "Unable to fit the lowest batch size of 1 for inference, try methods to reduce the gpu consumption"
                    ) from exception

    return inner_wrapper


def gc_cuda():
    """Gargage collect RAM & Torch (CUDA) memory."""
    gc.collect()
    if torch.cuda.is_available():
        ctypes.CDLL("libc.so.6").malloc_trim(0)
        torch.cuda.empty_cache()


def should_reduce_batch_size(exception):
    """Checks whether batch size can be reduced or not"""
    return (
        is_cuda_out_of_memory(exception)
        or is_cudnn_snafu(exception)
        or is_out_of_cpu_memory(exception)
    )


def is_cuda_out_of_memory(exception):
    """Checks for CUDA OOM Error"""
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "CUDA out of memory." in exception.args[0]
    )


def is_cudnn_snafu(exception):
    """For/because of https://github.com/pytorch/pytorch/issues/4107"""
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "cuDNN error: CUDNN_STATUS_NOT_SUPPORTED." in exception.args[0]
    )


def is_out_of_cpu_memory(exception):
    """Checks for CPU OOM Error"""
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "DefaultCPUAllocator: can't allocate memory" in exception.args[0]
    )


def get_gpu_information() -> Union[None, Tuple[int, float, float]]:
    """Get CUDA gpu related info if gpu exist

    Returns:
        Union[None, Tuple[int, float, float]]: total available gpus, total occupied memory gb, total available gpu memeory
        `None` if unable to get CUDA GPU related info
    """
    try:
        pynvml.nvmlInit()
        try:
            num_gpus = pynvml.nvmlDeviceGetCount()

            total_available_gpus = 0
            total_occupied_memory = 0
            total_memory = 0

            for index in range(num_gpus):
                handle = pynvml.nvmlDeviceGetHandleByIndex(index)
                memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)

                total_available_gpus += 1
                total_occupied_memory += memory_info.used
                total_memory += memory_info.total

            # Round occupied memory to the nearest GB
            total_occupied_memory_gb = math.ceil(total_occupied_memory / (1024**3))

            # Round total memory to the nearest GB
            total_memory_gb = math.ceil(total_memory / (1024**3))

            total_available_memory = total_memory_gb - total_occupied_memory_gb

            return (
                total_available_gpus,
                total_occupied_memory_gb,
                total_available_memory,
            )
        except Exception:
            exc_traceback = traceback.format_exc()
            print(f"Unable to get details of gpu - {exc_traceback}")
            raise
        finally:
            pynvml.nvmlShutdown()
    except BaseException:
        return None


def get_total_available_ram() -> int:
    """Get total available ram in GB

    Returns:
        float: available ram in GB
    """
    memory = psutil.virtual_memory()
    total_available_ram = memory.available

    # Round available RAM to the nearest GB
    total_available_ram_gb = math.ceil(total_available_ram / (1024**3))

    return total_available_ram_gb



# pylint: skip-file
"""
Monkey Patch `transformers` to add in generation techniques that are currently not supported by the `transformers` library.
Adds `tfs`, `top_a` & `generation_seed` to `transformers.GenerationConfig`

Implementation Reference of `tfs` & `top_a` from [oobabooga/text-generation-webui](https://github.com//blob/main/modules/sampler_hijack.py)
"""
import math
from typing import Union

import torch
import warnings
import numpy as np
import transformers
from transformers import LogitsWarper, StoppingCriteriaList
from transformers.generation.logits_process import (
    LogitNormalization,
    LogitsProcessor,
    LogitsProcessorList,
    TemperatureLogitsWarper,
)

from llmsearch.utils.model_utils import seed_everything


class TailFreeLogitsWarper(LogitsWarper):
    """[TFS sampling](https://www.trentonbricken.com/Tail-Free-Sampling/)"""

    def __init__(
        self,
        tfs: float,
        filter_value: float = -float("Inf"),
        min_tokens_to_keep: int = 1,
    ):
        tfs = float(tfs)
        if tfs < 0 or tfs > 1.0:
            raise ValueError(f"`tfs` has to be a float >= 0 and <= 1, but is {tfs}")
        self.tfs = tfs
        self.filter_value = filter_value
        self.min_tokens_to_keep = min_tokens_to_keep

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        sorted_logits, sorted_indices = torch.sort(scores, descending=True)
        probs = sorted_logits.softmax(dim=-1)

        # Compute second derivative normalized CDF
        d2 = probs.diff().diff().abs()
        normalized_d2 = d2 / d2.sum(dim=-1, keepdim=True)
        # compute cdf from pdf
        normalized_d2_cdf = normalized_d2.cumsum(dim=-1)

        # Remove tokens with CDF value above the threshold (token with 0 are kept)
        sorted_indices_to_remove = normalized_d2_cdf > self.tfs

        # Centre the distribution around the cutoff as in the original implementation of the algorithm
        sorted_indices_to_remove = torch.cat(
            (
                torch.zeros(scores.shape[0], 1, dtype=torch.bool, device=scores.device),
                sorted_indices_to_remove,
                torch.ones(scores.shape[0], 1, dtype=torch.bool, device=scores.device),
            ),
            dim=-1,
        )

        if self.min_tokens_to_keep > 1:
            # Keep at least min_tokens_to_keep
            sorted_indices_to_remove[..., : self.min_tokens_to_keep] = 0

        indices_to_remove = sorted_indices_to_remove.scatter(
            1, sorted_indices, sorted_indices_to_remove
        )
        scores = scores.masked_fill(indices_to_remove, self.filter_value)
        return scores


class TopALogitsWarper(LogitsWarper):
    """[Top-A sampling](https://github.com/BlinkDL/RWKV-LM/tree/4cb363e5aa31978d801a47bc89d28e927ab6912e#the-top-a-sampling-method.) Introduced in RWKV-LM repo."""

    def __init__(
        self,
        top_a: float,
        filter_value: float = -float("Inf"),
        min_tokens_to_keep: int = 1,
    ):
        top_a = float(top_a)
        if top_a < 0 or top_a > 1.0:
            raise ValueError(f"`top_a` has to be a float >= 0 and <= 1, but is {top_a}")
        self.top_a = top_a
        self.filter_value = filter_value
        self.min_tokens_to_keep = min_tokens_to_keep

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        sorted_logits, sorted_indices = torch.sort(scores, descending=True)
        probs = sorted_logits.softmax(dim=-1)

        # Remove tokens with probability less than top_a*(max(probs))^2 (token with 0 are kept)
        probs_max = probs[..., 0, None]
        sorted_indices_to_remove = probs < probs_max * probs_max * self.top_a

        if self.min_tokens_to_keep > 1:
            # Keep at least min_tokens_to_keep
            sorted_indices_to_remove[..., : self.min_tokens_to_keep] = 0

        indices_to_remove = sorted_indices_to_remove.scatter(
            1, sorted_indices, sorted_indices_to_remove
        )
        scores = scores.masked_fill(indices_to_remove, self.filter_value)
        return scores


class MirostatLogitsWarper(LogitsWarper):
    """[Mirostat sampling](https://arxiv.org/pdf/2007.14966.pdf) - Currently not supported, needs work to work with a batch of sequences.

    raises `NotImplementedError` if it gets called
    """

    def __init__(
        self,
        mirostat_mode: int,
        mirostat_tau: float,
        mirostat_eta: float,
        filter_value: float = -float("Inf"),
        min_tokens_to_keep: int = 1,
        generation_seed: Union[int, None] = None,
    ):
        raise NotImplementedError(
            f"Found `mirostat_mode` - `{mirostat_mode}` in generation parameters, mirostat is not fully implemented yet, Try a different generation method"
        )
        if mirostat_mode not in [2]:
            raise ValueError(
                f"`mirostat` has to be a an integer 2, but is {mirostat_mode}"
            )
        self.mirostat_mode = mirostat_mode
        # learning rate
        self.mirostat_eta = mirostat_eta
        # target surprise value
        self.mirostat_tau = mirostat_tau
        self.filter_value = filter_value
        self.min_tokens_to_keep = min_tokens_to_keep
        # max cross entropy/surprise
        self.mu = 2 * self.mirostat_tau
        self.batch_init_done = False
        self.generation_seed = generation_seed

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        """
        Batch Mode
        - current generation step is n
        - This method is called for each generation step, shape of inputs_ids starts with (batch_size, 1), then (batch_size, 2)
        it contains the previous token ID

        Args:
            input_ids (torch.LongTensor): (batch_size, step_idx) step_idx is the the current generation step
            scores (torch.FloatTensor): (batch_size, vocab_size) scores for the current token

        Returns:
            torch.FloatTensor: _description_
        """
        if not self.batch_init_done:
            # Initialize state when first time the function is called
            self.mu = torch.full(size=(input_ids.shape[0],), fill_value=float(self.mu))
            self.batch_init_done = True

        # (batch_size, vocab_size), (batch_size, vocab_size)
        sorted_logits, sorted_indices = torch.sort(scores, descending=True, dim=-1)

        # calculate probs from logits - (batch_size, vocab_size)
        prob_original = torch.softmax(sorted_logits, dim=-1)

        # Get candidate logits for each sample
        sorted_logits_batch = (
            []
        )  # (list of logits for each row, each list contains candidate logits fitting the criteria)
        # Go through each row's top probabilities
        j = None
        for i, row in enumerate(prob_original):
            # Truncate/discard the words with surprise values greater than mu
            # as you go below, surprise/cross entropy value increases
            row_logits = []
            for j, candidate in enumerate(row):
                if candidate > 0 and -math.log2(candidate) > self.mu[i]:
                    # If there is a single candidate only take that
                    if j == 0:
                        row_logits.append(sorted_logits[i][:1])
                    # If there is more than one, take all but this
                    else:
                        row_logits.append(sorted_logits[i][:j])
                    break
            row_logit_tensor = torch.cat(row_logits, dim=0)
            sorted_logits_batch.append(row_logit_tensor)

        prev_indices = []
        final_scores = []

        # iterate through candidates of each sample
        for i, row in enumerate(sorted_logits_batch):
            softmaxed_candidates = torch.softmax(row.reshape(-1), dim=0)

            # pick a candidate index from all of the selected candidates
            if self.generation_seed:
                seed_everything(seed=self.generation_seed)
            prev_index = torch.multinomial(
                softmaxed_candidates, num_samples=1, replacement=True
            )
            prev_indices.append(prev_index)

            observed_surprise_value = -math.log2(softmaxed_candidates[prev_index])

            # calculate error based on observed surprise value and target surprise value
            e = observed_surprise_value - self.mirostat_tau

            # update mu using learning rate and error

            self.mu[i] -= self.mirostat_eta * e

            # shape - (vocab_size,)
            sorted_indices_to_remove = torch.ones_like(scores[0], dtype=torch.bool)
            # set the token that is picked to False
            sorted_indices_to_remove[prev_index] = False

            # shape - (vocab_size,)
            # convert the mask to be able to map to the scores(unsorted version)
            indices_to_remove = sorted_indices_to_remove.unsqueeze(0).scatter(
                1, sorted_indices[i].unsqueeze(0), sorted_indices_to_remove.unsqueeze(0)
            )
            # mask all other tokens
            scores[i] = scores[i].masked_fill(indices_to_remove, self.filter_value)
        return scores

    def __call_(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        """Non-batch mode"""
        # This happens step by step, each token is passed in, then logit is calculated
        logits = scores[0]
        sorted_logits, sorted_indices = torch.sort(logits, descending=True, dim=-1)
        prob_original = torch.softmax(sorted_logits, dim=-1)  # candidates

        # Truncate the words with surprise values greater than mu
        # as you go below, surprise/cross entropy value increases
        for i, candidate in enumerate(prob_original):
            if candidate > 0 and -math.log2(candidate) > self.mu:
                if i == 0:
                    sorted_logits = sorted_logits[:1]
                else:
                    sorted_logits = sorted_logits[:i]
                break

        # Normalize the probabilities of the remaining words
        prob_topk = torch.softmax(sorted_logits, dim=0)

        if self.generation_seed:
            seed_everything(seed=self.generation_seed)
        prev_i = torch.multinomial(prob_topk, num_samples=1, replacement=True)

        observed_surprise = -math.log2(prob_topk[prev_i])
        self.e = observed_surprise - self.mirostat_tau

        # Update mu using the learning rate and error
        self.mu -= self.mirostat_eta * self.e

        sorted_indices_to_remove = torch.ones_like(scores[0], dtype=torch.bool)
        sorted_indices_to_remove[prev_i] = False

        indices_to_remove = sorted_indices_to_remove.unsqueeze(0).scatter(
            1, sorted_indices.unsqueeze(0), sorted_indices_to_remove.unsqueeze(0)
        )
        random_tensor = torch.arange(1, len(indices_to_remove[0]) + 1).to("mps")
        scores = scores.masked_fill(indices_to_remove, self.filter_value)
        return scores


def get_logits_warper_patch(self, generation_config):
    """Patch to add in new generation techniques to `transformers.GenerationMixin._get_logits_warper`
    Adds in `tfs`, `top_a` and `mirostat` (not supported currently by `llmsearch`) to the generation techniques
    """
    warpers = self._get_logits_warper_old(generation_config)
    warpers_to_add = LogitsProcessorList()
    min_tokens_to_keep = 2 if generation_config.num_beams > 1 else 1
    # only apply mirostat if mode is 2
    if (
        generation_config.mirostat_mode is not None
        and generation_config.mirostat_mode == 2
    ):
        warpers_to_add.append(
            MirostatLogitsWarper(
                mirostat_mode=generation_config.mirostat_mode,
                mirostat_eta=generation_config.mirostat_eta,
                mirostat_tau=generation_config.mirostat_tau,
                min_tokens_to_keep=min_tokens_to_keep,
                generation_seed=generation_config.generation_seed,
            )
        )
        # We need to disable samplers other than temperature
        for warper in warpers:
            if not isinstance(warper, TemperatureLogitsWarper):
                warpers.remove(warper)
    else:
        if generation_config.tfs is not None and 0.0 <= generation_config.tfs <= 1.0:
            warpers_to_add.append(
                TailFreeLogitsWarper(
                    tfs=generation_config.tfs, min_tokens_to_keep=min_tokens_to_keep
                )
            )
        if (
            generation_config.top_a is not None
            and 0.0 <= generation_config.top_a <= 1.0
        ):
            warpers_to_add.append(
                TopALogitsWarper(
                    top_a=generation_config.top_a, min_tokens_to_keep=min_tokens_to_keep
                )
            )

    if warpers and isinstance(warpers[-1], LogitNormalization):
        warpers = warpers[:-1] + warpers_to_add + [warpers[-1]]
    else:
        warpers += warpers_to_add

    return warpers


def generation_config_init_patch(self, **kwargs):
    """Patch to add in new generation parameters to `transformers.GenerationConfig`"""
    self.__init___old(**kwargs)
    self.tfs = kwargs.pop("tfs", None)
    self.top_a = kwargs.pop("top_a", None)
    self.generation_seed = kwargs.pop("generation_seed", None)

    # TODO : Implement Mirostat, batching requires state
    self.mirostat_mode = kwargs.pop("mirostat_mode", 0)
    self.mirostat_eta = kwargs.pop("mirostat_eta", 0.1)
    self.mirostat_tau = kwargs.pop("mirostat_tau", 5)


def hijack_samplers():
    """Patches generation methods to add in new generation techniques

    - `transformers.GenerationMixin._get_logits_warper` to add in new generation techniques, Older version of the method is stored in `_get_logits_warper_old`
    - `transformers.GenerationConfig.__init__` to add in new generation parameters, Older version of the method is stored in `__init___old`

    """
    transformers.GenerationMixin._get_logits_warper_old = (
        transformers.GenerationMixin._get_logits_warper
    )
    transformers.GenerationMixin._get_logits_warper = get_logits_warper_patch

    transformers.GenerationConfig.__init___old = transformers.GenerationConfig.__init__
    transformers.GenerationConfig.__init__ = generation_config_init_patch


def __sklearn_clone__(self):
    return self


StoppingCriteriaList.__sklearn_clone__ = __sklearn_clone__



"""Model downloader - https://github.com/oobabooga/text-generation-webui/blob/main/download-model.py"""
import argparse
import base64
import datetime
import hashlib
import json
import os
import re
import sys
from pathlib import Path

import requests
import tqdm
from requests.adapters import HTTPAdapter
from tqdm.contrib.concurrent import thread_map

base = "https://huggingface.co"


class ModelDownloader:
    def __init__(self, max_retries=5):
        self.session = requests.Session()
        if max_retries:
            self.session.mount(
                "https://cdn-lfs.huggingface.co", HTTPAdapter(max_retries=max_retries)
            )
            self.session.mount(
                "https://huggingface.co", HTTPAdapter(max_retries=max_retries)
            )
        if os.getenv("HF_USER") is not None and os.getenv("HF_PASS") is not None:
            self.session.auth = (os.getenv("HF_USER"), os.getenv("HF_PASS"))
        if os.getenv("HF_TOKEN") is not None:
            self.session.headers = {"authorization": f'Bearer {os.getenv("HF_TOKEN")}'}

    def sanitize_model_and_branch_names(self, model, branch):
        if model[-1] == "/":
            model = model[:-1]

        if model.startswith(base + "/"):
            model = model[len(base) + 1 :]

        model_parts = model.split(":")
        model = model_parts[0] if len(model_parts) > 0 else model
        branch = model_parts[1] if len(model_parts) > 1 else branch

        if branch is None:
            branch = "main"
        else:
            pattern = re.compile(r"^[a-zA-Z0-9._-]+$")
            if not pattern.match(branch):
                raise ValueError(
                    "Invalid branch name. Only alphanumeric characters, period, underscore and dash are allowed."
                )

        return model, branch

    def get_download_links_from_huggingface(
        self, model, branch, text_only=False, specific_file=None
    ):
        page = f"/api/models/{model}/tree/{branch}"
        cursor = b""

        links = []
        sha256 = []
        classifications = []
        has_pytorch = False
        has_pt = False
        has_gguf = False
        has_safetensors = False
        is_lora = False
        while True:
            url = f"{base}{page}" + (f"?cursor={cursor.decode()}" if cursor else "")
            r = self.session.get(url, timeout=10)
            r.raise_for_status()
            content = r.content

            dict = json.loads(content)
            if len(dict) == 0:
                break

            for i in range(len(dict)):
                fname = dict[i]["path"]
                if specific_file not in [None, ""] and fname != specific_file:
                    continue

                if not is_lora and fname.endswith(
                    ("adapter_config.json", "adapter_model.bin")
                ):
                    is_lora = True

                is_pytorch = re.match(r"(pytorch|adapter|gptq)_model.*\.bin", fname)
                is_safetensors = re.match(r".*\.safetensors", fname)
                is_pt = re.match(r".*\.pt", fname)
                is_gguf = re.match(r".*\.gguf", fname)
                is_tiktoken = re.match(r".*\.tiktoken", fname)
                is_tokenizer = (
                    re.match(r"(tokenizer|ice|spiece).*\.model", fname) or is_tiktoken
                )
                is_text = re.match(r".*\.(txt|json|py|md)", fname) or is_tokenizer
                if any(
                    (is_pytorch, is_safetensors, is_pt, is_gguf, is_tokenizer, is_text)
                ):
                    if "lfs" in dict[i]:
                        sha256.append([fname, dict[i]["lfs"]["oid"]])

                    if is_text:
                        links.append(
                            f"https://huggingface.co/{model}/resolve/{branch}/{fname}"
                        )
                        classifications.append("text")
                        continue

                    if not text_only:
                        links.append(
                            f"https://huggingface.co/{model}/resolve/{branch}/{fname}"
                        )
                        if is_safetensors:
                            has_safetensors = True
                            classifications.append("safetensors")
                        elif is_pytorch:
                            has_pytorch = True
                            classifications.append("pytorch")
                        elif is_pt:
                            has_pt = True
                            classifications.append("pt")
                        elif is_gguf:
                            has_gguf = True
                            classifications.append("gguf")

            cursor = (
                base64.b64encode(f'{{"file_name":"{dict[-1]["path"]}"}}'.encode())
                + b":50"
            )
            cursor = base64.b64encode(cursor)
            cursor = cursor.replace(b"=", b"%3D")

        # If both pytorch and safetensors are available, download safetensors only
        if (has_pytorch or has_pt) and has_safetensors:
            for i in range(len(classifications) - 1, -1, -1):
                if classifications[i] in ["pytorch", "pt"]:
                    links.pop(i)

        # For GGUF, try to download only the Q4_K_M if no specific file is specified.
        # If not present, exclude all GGUFs, as that's likely a repository with both
        # GGUF and fp16 files.
        if has_gguf and specific_file is None:
            has_q4km = False
            for i in range(len(classifications) - 1, -1, -1):
                if "q4_k_m" in links[i].lower():
                    has_q4km = True

            if has_q4km:
                for i in range(len(classifications) - 1, -1, -1):
                    if "q4_k_m" not in links[i].lower():
                        links.pop(i)
            else:
                for i in range(len(classifications) - 1, -1, -1):
                    if links[i].lower().endswith(".gguf"):
                        links.pop(i)

        is_llamacpp = has_gguf and specific_file is not None
        return links, sha256, is_lora, is_llamacpp

    def get_output_folder(
        self, model, branch, is_lora, is_llamacpp=False, base_folder=None
    ):
        if base_folder is None:
            base_folder = "models" if not is_lora else "loras"

        # If the model is of type GGUF, save directly in the base_folder
        if is_llamacpp:
            return Path(base_folder)

        output_folder = f"{'_'.join(model.split('/')[-2:])}"
        if branch != "main":
            output_folder += f"_{branch}"

        output_folder = Path(base_folder) / output_folder
        return output_folder

    def get_single_file(self, url, output_folder, start_from_scratch=False):
        filename = Path(url.rsplit("/", 1)[1])
        output_path = output_folder / filename
        headers = {}
        mode = "wb"
        if output_path.exists() and not start_from_scratch:

            # Check if the file has already been downloaded completely
            r = self.session.get(url, stream=True, timeout=10)
            total_size = int(r.headers.get("content-length", 0))
            if output_path.stat().st_size >= total_size:
                return

            # Otherwise, resume the download from where it left off
            headers = {"Range": f"bytes={output_path.stat().st_size}-"}
            mode = "ab"

        with self.session.get(url, stream=True, headers=headers, timeout=10) as r:
            r.raise_for_status()  # Do not continue the download if the request was unsuccessful
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024 * 1024  # 1MB

            tqdm_kwargs = {
                "total": total_size,
                "unit": "iB",
                "unit_scale": True,
                "bar_format": "{l_bar}{bar}| {n_fmt:6}/{total_fmt:6} {rate_fmt:6}",
            }

            if "COLAB_GPU" in os.environ:
                tqdm_kwargs.update({"position": 0, "leave": True})

            with open(output_path, mode) as f:
                with tqdm.tqdm(**tqdm_kwargs) as t:
                    count = 0
                    for data in r.iter_content(block_size):
                        t.update(len(data))
                        f.write(data)
                        if total_size != 0 and self.progress_bar is not None:
                            count += len(data)
                            self.progress_bar(
                                float(count) / float(total_size), f"{filename}"
                            )

    def start_download_threads(
        self, file_list, output_folder, start_from_scratch=False, threads=4
    ):
        thread_map(
            lambda url: self.get_single_file(
                url, output_folder, start_from_scratch=start_from_scratch
            ),
            file_list,
            max_workers=threads,
            disable=True,
        )

    def download_model_files(
        self,
        model,
        branch,
        links,
        sha256,
        output_folder,
        progress_bar=None,
        start_from_scratch=False,
        threads=4,
        specific_file=None,
        is_llamacpp=False,
    ):
        self.progress_bar = progress_bar

        # Create the folder and writing the metadata
        output_folder.mkdir(parents=True, exist_ok=True)

        if not is_llamacpp:
            metadata = (
                f"url: https://huggingface.co/{model}\n"
                f"branch: {branch}\n"
                f'download date: {datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n'
            )

            sha256_str = "\n".join([f"    {item[1]} {item[0]}" for item in sha256])
            if sha256_str:
                metadata += f"sha256sum:\n{sha256_str}"

            metadata += "\n"
            (output_folder / "huggingface-metadata.txt").write_text(metadata)

        if specific_file:
            print(f"Downloading {specific_file} to {output_folder}")
        else:
            print(f"Downloading the model to {output_folder}")

        self.start_download_threads(
            links, output_folder, start_from_scratch=start_from_scratch, threads=threads
        )

    def check_model_files(self, model, branch, links, sha256, output_folder):
        # Validate the checksums
        validated = True
        for i in range(len(sha256)):
            fpath = output_folder / sha256[i][0]

            if not fpath.exists():
                print(f"The following file is missing: {fpath}")
                validated = False
                continue

            with open(output_folder / sha256[i][0], "rb") as f:
                bytes = f.read()
                file_hash = hashlib.sha256(bytes).hexdigest()
                if file_hash != sha256[i][1]:
                    print(f"Checksum failed: {sha256[i][0]}  {sha256[i][1]}")
                    validated = False
                else:
                    print(f"Checksum validated: {sha256[i][0]}  {sha256[i][1]}")

        if validated:
            print("[+] Validated checksums of all model files!")
        else:
            print(
                "[-] Invalid checksums. Rerun download-model.py with the --clean flag."
            )


def download_model_from_hf(model_id, save_dir, branch="main", threads=4, max_retries=5):
    """Download model from HF
         >>> download_model_from_hf(model_id, save_dir=temp_model_dir, branch="main")
    """

    downloader = ModelDownloader(max_retries=max_retries)

    model_id, branch = downloader.sanitize_model_and_branch_names(model_id, branch)
    (
        links,
        sha256,
        is_lora,
        is_llamacpp,
    ) = downloader.get_download_links_from_huggingface(
        model_id, branch, text_only=False, specific_file=None
    )

    output_folder = downloader.get_output_folder(
        model_id, branch, is_lora, is_llamacpp=is_llamacpp, base_folder=save_dir
    )

    if Path(output_folder).exists():
        print(f"Model already exists in {output_folder}. Checking the model files...")
        downloader.check_model_files(model_id, branch, links, sha256, output_folder)
        return output_folder

    downloader.download_model_files(
        model_id,
        branch,
        links,
        sha256,
        output_folder,
        specific_file=None,
        threads=threads,
        is_llamacpp=is_llamacpp,
    )

    return output_folder



"""
Main Tuner Module containing `LLMEstimatorWrapper` Class & `Tuner` Class for scikit-learn
"""

import random
from operator import itemgetter
from typing import List, Union, Tuple, Dict, Callable

import numpy as np
from torch import nn
from datasets import Dataset
from sklearn.base import BaseEstimator
from sklearn.metrics import make_scorer
from transformers import AutoTokenizer

from llmsearch.utils.model_utils import run_inference
from llmsearch.utils.mem_utils import cache
from llmsearch.utils.common_utils import process_dataset_with_map

from llmsearch.utils.logging_utils import get_logger

logger = get_logger(__name__)


class LLMEstimatorWrapper(BaseEstimator):
    """Estimator Wrapper Class that abstracts a Model for compatibility with scikit-learn"""

    def __init__(
        self,
        model: nn.Module,
        tokenizer: AutoTokenizer,
        scorer: Callable,
        device: str,
        tokenizer_encode_args: Dict,
        tokenizer_decode_args: Dict,
        batch_size: int,
        callbacks_after_inference: List[Callable] = None,
        disable_batch_size_cache: bool = False,
        pred_function: Union[Callable, None] = None,
        is_encoder_decoder: bool = False,
        disable_generation_param_checks: bool = False,
        output_preproc: Callable = lambda x: x.strip(),
        **kwargs,
    ):
        """Initializes the Estimator, All `**kwargs` are assumed to be generation parameters for the model.

        Args:
            model (nn.Module): model that has a `.generate` method
            tokenizer (AutoTokenizer): tokenizer for the input
            scorer (Callable): A function that has this signature - `(y_true: List, y_pred: List) -> float` , takes in ground truth and predictions are returns a metric to optimize on
            device (str): device to run inference on, eg - `cuda:0`
            tokenizer_encode_args (Dict): Encoding key value arguments for the `tokenizer`
            tokenizer_decode_args (Dict): Decoding key value arguments for the `tokenizer`
            batch_size (int): batch_size to run inference with, this gets dynamically halfed if the inference function encounters OOM errors
            callbacks_after_inference (List[Callable], optional): Callbacks to run after each inference. Useful for stopping criteria in generation. Defaults to `None`.
            disable_batch_size_cache (bool): If `True` for each cross validation run, the pre-defined `batch_size` is used, this could lead to wasted computation time if OOM is raised by the inference function, Defaults to `False`.
            pred_function (Union[Callable, None], optional): Override inference function if present. Defaults to `None`.
            is_encoder_decoder (bool, optional): whether the model is an encoder-decoder model, `False` if not. Defaults to `False`.
            disable_generation_param_checks (bool, optional): Disables the custom generation parameter checks, this does a sanity check of the parameters & produces warnings before doing generation. Defaults to `False`.
            output_preproc (Callable, optional): Post processing function for the output, by default it strips the output. Defaults to `lambda x : x.strip()`.
        """
        self.model = model
        self.tokenizer = tokenizer
        self.scorer = scorer
        self.device = device
        self.tokenizer_encode_args = tokenizer_encode_args
        self.tokenizer_decode_args = tokenizer_decode_args
        self.batch_size = batch_size
        # stores the optimal batch size for a particular configuration if disable_batch_size_cache is `False`
        self._optimal_batch_size = batch_size
        self.callbacks_after_inference = (
            [] if callbacks_after_inference is None else callbacks_after_inference
        )
        self.disable_batch_size_cache = disable_batch_size_cache
        self.pred_function = pred_function
        self.is_encoder_decoder = is_encoder_decoder
        self.disable_generation_param_checks = disable_generation_param_checks
        self.output_preproc = output_preproc

        # Set generation params, All kwargs are assumed to be generation params
        for k, v in kwargs.items():
            self.__setattr__(k, v)

        # Stores the names of the keys passed in
        self._model_generation_param_keys = list(kwargs.keys())
        logger.debug(
            "Initializing new estimator with generation parameters - %s", kwargs
        )

    @property
    def optimal_batch_size(self):
        """Stores the optimal batch size for a particular configuration if disable_batch_size_cache is `False`"""
        return self._optimal_batch_size

    @property
    def model_generation_param_keys(self):
        """Stores the names of the keys passed in"""
        return self._model_generation_param_keys

    def fit(self, *args, **kwargs) -> BaseEstimator:  # pylint: disable=unused-argument
        """Dummy fit function which does not actually do anything :)

        Returns:
            self: returns self
        """
        self.is_fitted_ = True
        return self

    def predict(self, X: List) -> List:
        """Performs prediction on list of inputs `X`

        Args:
            X (List): list of inputs

        Returns:
            List: output produced by the model for each sample in `X`
        """
        # Gets the value of generation params using which prediction will run, when running any kind of search this is set via `.set_params`
        model_generation_params = {
            attr: getattr(self, attr) for attr in self._model_generation_param_keys
        }
        # Override inference function if present
        if self.pred_function:
            return self.pred_function(self, X, model_generation_params)
        # Perform inference on input data
        output, self._optimal_batch_size = run_inference(
            model=self.model,
            tokenizer=self.tokenizer,
            is_encoder_decoder=self.is_encoder_decoder,
            batch_size=self.batch_size,
            disable_batch_size_cache=self.disable_batch_size_cache,
            device=self.device,
            model_inputs=X,
            tokenizer_encode_args=self.tokenizer_encode_args,
            generation_args=model_generation_params,
            disable_generation_param_checks=self.disable_generation_param_checks,
            tokenizer_decode_args=self.tokenizer_decode_args,
            return_optimal_batch_size=True,
            callbacks=self.callbacks_after_inference,
            output_preproc=self.output_preproc,
        )
        return output

    def __sklearn_clone__(self) -> BaseEstimator:
        """Returns the same estimator
        Only relevant from sklearn >= 1.3

        - This functions returns the same estimator, as there are no parameters to specifically "fit"
        - This is done to avoid OOM errors for larger models and prevent creating a copy of the model weights, this does not affect the hyperparameter search in any way

        Returns:
            BaseEstimator: self
        """
        _ = self._get_model_generation_params()
        logger.debug("Attributes before cloning estimator - %s", _)
        return self

    def get_params(self, deep: bool = True):
        """
        Gets all of the parameters defined when calling `__init__` + generation parameters if set

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = {}

        for key, value in vars(self).items():
            # Ignore any private/protected variables
            if not (key.startswith("__") or key.startswith("_")):
                out[key] = value
        return out

    def set_params(self, **params):
        """Sets the generation parameters of this estimator

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : estimator instance
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self

        # set the keys each time this method is called
        self._model_generation_param_keys = list(params.keys())

        for key, value in params.items():
            setattr(self, key, value)

        _ = self._get_model_generation_params()

        return self

    def _get_model_generation_params(self):
        model_generation_params = {
            attr: getattr(self, attr) for attr in self._model_generation_param_keys
        }
        return model_generation_params


class Tuner:
    """Tuner Class which drives the search for the generation hyperparameters"""

    def __init__(
        self,
        model: nn.Module,
        tokenizer: AutoTokenizer,
        dataset: Dataset,
        column_mapping: Dict[str, list],
        scorer: Callable,
        device: str,
        prompt_template: str = None,
        sample_preprocessor: Callable = None,
        tokenizer_encode_args: Dict = None,
        tokenizer_decode_args: Dict = None,
        batch_size: int = 16,
        output_preproc: Callable = lambda x: x.strip(),
        callbacks_after_inference: List[Callable] = None,
        is_encoder_decoder: bool = False,
        greater_is_better: bool = True,
        seed: int = 42,
        disable_batch_size_cache: bool = False,
        disable_generation_param_checks: bool = False,
        sample_ratio: float = 0.3,
        tokenizer_max_length_quantile: float = 0.9,
        custom_pred_function=None,
    ):
        """
        Initializes the Tuner Class, populates the dataset with `_X` & `_y` keys for the model to use.

        Args:
            model (nn.Module): model that has a `.generate` method
            tokenizer (AutoTokenizer): tokenizer for the input
            dataset (Dataset): dataset to perform search on
            column_mapping (Dict[str, list]): should contain `input_cols` & `eval_cols` keys, `input_cols` should contain the columns to be used in the `prompt_template` & `eval_cols` should contain the columns to be used in the `scorer`,all eval_columns will be passed in as a dict as the second argument to the `scorer` function, eg - `{'input_cols' : ["question"], 'eval_cols' : ['answer']}`
            scorer (Callable): A function that has this signature - `(y_true: List, y_pred: List) -> float` , takes in ground truth and predictions are returns a metric to optimize on, `eval_cols` in `column_mapping` are passed in as the second argument as a `List[Dict]`
            device (str): device to run inference on, eg - `cuda:0`
            prompt_template (str): template with placeholders for `input_cols` from the `column_mapping` argument, eg - `"Question : How many days are there in a year?\\nAnswer : 365\\n\\nQuestion : {question}\\nAnswer : "`, not used when `sample_preprocessor` is not `None`
            sample_preprocessor (Callable): Preprocessor function for a single example from the `dataset`, should have the signature - `(tokenizer, **kwargs) -> str`, where key word arguments are the columns from `input_cols` & `eval_cols` in `column_mapping`, not used when `prompt_template` is not `None`
            tokenizer_encode_args (Dict, optional): Encoding key value arguments for the `tokenizer`. If `None` it's initialized using the `get_default_input_tokenizer_kwargs` method. Defaults to `None`.
            tokenizer_decode_args (Dict, optional): Decoding key value arguments for the `tokenizer`. Defaults to `{'skip_special_tokens' : True}`.
            batch_size (int, optional): batch_size to run inference with, this gets dynamically halfed if the inference function encounters OOM errors. Defaults to `16`.
            output_preproc (Callable, optional): Post processing function for the output, by default it strips the output. Defaults to `lambda x : x.strip()`.
            callbacks_after_inference (List[Callable], optional): Callbacks to run after each inference. Useful for stopping criteria in generation. Defaults to `None`.
            is_encoder_decoder (bool, optional): whether the model is an encoder-decoder model, `False` if not. Defaults to `False`.
            greater_is_better (bool, optional): whether the metric to optimize on is greater the better. Defaults to `True`.
            seed (int, optional): seed for reproducibility. Defaults to `42`.
            disable_batch_size_cache (bool, optional): If `True` for each cross validation run, the pre-defined `batch_size` is used, this could lead to wasted computation time if OOM is raised by the inference function. Defaults to `False`.
            disable_generation_param_checks (bool, optional): Disables the custom generation parameter checks, this does a sanity check of the parameters & produces warnings before doing generation. Defaults to `False`.
            sample_ratio (float, optional): Sampling Ratio of `dataset` to find the ideal values for padding and truncation. Argument is invalid if `tokenizer_encode_args` is not `None`. Defaults to `0.3`.
            tokenizer_max_length_quantile (float, optional): percentile to find a value for `max_length` based on the dataset. Defaults to `0.9`.
            custom_pred_function (Union[Callable, None], optional): Override inference function if present. Defaults to `None`. Should take in two parameters - model inputs (`List[str]`) and model generation parameters (`Dict`) and return a `List[str]` of outputs overrides `model_utils.run_inference`
        """
        self.tokenizer = tokenizer
        self.prompt_template = prompt_template
        self.sample_preprocessor = sample_preprocessor
        self.column_mapping = column_mapping
        self.input_cols = column_mapping["input_cols"]
        self.eval_cols = column_mapping["eval_cols"]

        assert (
            self.prompt_template or self.sample_preprocessor
        ), "`prompt_template` or `sample_preprocessor` should be provided, got `None` for both"
        assert not (
            self.prompt_template and self.sample_preprocessor
        ), "Only one of `prompt_template` or `sample_preprocessor` should be provided"

        self.dataset = self.preprocess_dataset(dataset=dataset)
        self.device = device
        self.score_func = scorer
        self.scorer = make_scorer(
            score_func=scorer, greater_is_better=greater_is_better
        )
        self.seed = seed
        self.disable_batch_size_cache = disable_batch_size_cache
        self.disable_generation_param_checks = disable_generation_param_checks

        # TODO - to check
        self.sample_ratio = sample_ratio
        self.tokenizer_max_length_quantile = tokenizer_max_length_quantile

        # Get encoding arguments for the `tokenizer` if not passed in
        self.tokenizer_encode_args = self.get_default_tokenizer_encode_args(
            sample_ratio=sample_ratio,
            tokenizer_encode_args=tokenizer_encode_args,
        )
        self.tokenizer_decode_args = (
            tokenizer_decode_args
            if tokenizer_decode_args
            else {"skip_special_tokens": True}
        )
        # Initialize the model estimator
        self.estimator = LLMEstimatorWrapper(
            model=model,
            tokenizer=self.tokenizer,
            scorer=self.scorer,
            device=self.device,
            tokenizer_encode_args=self.tokenizer_encode_args,
            tokenizer_decode_args=self.tokenizer_decode_args,
            batch_size=batch_size,
            callbacks_after_inference=callbacks_after_inference,
            disable_batch_size_cache=disable_batch_size_cache,
            pred_function=custom_pred_function,
            is_encoder_decoder=is_encoder_decoder,
            disable_generation_param_checks=disable_generation_param_checks,
            output_preproc=output_preproc,
        )
        # reset cache on every init
        cache.empty_cache()

    def preprocess_dataset(self, dataset: Dataset) -> Dataset:
        """Dataset preprocessor, preprocesses using the `prompt_template` or `sample_preprocessor` function.\n
        `self.prompt_template` - Useful for already processed datasets(text can be directly fed into the model)\n
        `self.sample_preprocessor` - Useful for datasets that need to be preprocessed(converting into chat format) before feeding into the model\n

        Adds `_X` & `_y` keys to the dataset

        **Note** : datasets.map is not used and traditional map has been used to map the dataset, as datasets.map has memory related issue. TODO :Issue to be raised in the datasets repo.

        Args:
            dataset (Dataset): dataset to preprocess
        """
        if self.prompt_template:
            processed_dataset = dataset.map(
                lambda sample: {
                    "_X": self.prompt_template.format(
                        **{
                            input_col: sample[input_col]
                            for input_col in self.input_cols
                        }
                    ),
                    "_y": {eval_col: sample[eval_col] for eval_col in self.eval_cols},
                }
            )
        elif self.sample_preprocessor:
                processed_dataset = process_dataset_with_map(
                    dataset=dataset,
                    sample_preprocessor=self.sample_preprocessor,
                    tokenizer=self.tokenizer,
                    input_cols=self.input_cols,
                    eval_cols=self.eval_cols,
                )
        return processed_dataset

    def get_default_tokenizer_encode_args(
        self,
        sample_ratio: float,
        tokenizer_encode_args: Union[Dict, None],
    ):
        """Get default input tokenizer arguments using the dataset, Sets `padding` & `truncation` to `True` and calculate `max_length` as tokenizer arguments.

        Args:
            sample_ratio (float): Sampling Ratio of `dataset` to find the ideal values for padding and truncation. Argument is invalid if `tokenizer_encode_args` is not `None`.
            tokenizer_length_percentile (float): percentile to find a value for `max_length` based on the dataset.
            tokenizer_kwargs (Union[Dict, None]): Encoding key value arguments for the `tokenizer`. Returns the same value if this is not `None`. Defaults to `None`.
        """
        # TODO - Check functionality of this method
        _quantiles_to_calc = [0.80, 0.85, 0.90, 0.92, 0.94, 0.95, 0.97, 0.99]

        if tokenizer_encode_args:
            return tokenizer_encode_args

        tokenizer_encode_args = {
            "padding": "longest",
            "truncation": True,
            "add_special_tokens": False,
        }
        logger.info(
            "Computing tokenizer encoding arguments using a sample of the dataset..."
        )
        sample_size = int(len(self.dataset["_X"]) * sample_ratio)
        random.seed(self.seed)
        sample_indexes = random.sample(range(0, len(self.dataset["_X"])), sample_size)
        get_items = itemgetter(*sample_indexes)
        X = get_items(self.dataset["_X"])
        # Calculate max_length
        max_length_at_quantile = self.get_value_at_quantile(
            input_list=X, quantile=self.tokenizer_max_length_quantile
        )

        # TODO : will max length and above config work?
        tokenizer_encode_args["max_length"] = min(
            max_length_at_quantile, self.tokenizer.model_max_length
        )
        logger.debug(
            "Computed max length at quantile - %d, tokenizer model max length - %d",
            max_length_at_quantile,
            tokenizer_encode_args["max_length"],
        )

        logger.info(
            "Setting tokenizer encoding arguments to - %s", tokenizer_encode_args
        )
        return tokenizer_encode_args

    def get_score(
        self, generation_args: Dict, dataset: Union[Dataset, Dict] = None
    ) -> Tuple[float, List]:
        """Evaluate the score function on a dataset or the initialized dataset using some generation arguments for the model. If `dataset` is `None` the initialized dataset is used, else the `dataset` is preprocessed(`_X` & `_y` are populated) and used.

        Args:
            generation_args (Dict): generation kwargs to perform inference
            dataset (Union[Dataset, Dict], optional): dataset to perform inference on. Defaults to None.

        Returns:
            Tuple[float, List]: score, predictions
        """
        dataset_to_evaluate = dataset if dataset else self.dataset
        if not dataset:
            dataset_to_evaluate = self.dataset
        else:
            dataset_to_evaluate = self.preprocess_dataset(dataset=dataset)
        y_pred = run_inference(
            model=self.estimator.model,
            tokenizer=self.estimator.tokenizer,
            is_encoder_decoder=self.estimator.is_encoder_decoder,
            batch_size=self.estimator.optimal_batch_size,
            device=self.device,
            model_inputs=dataset_to_evaluate["_X"],
            tokenizer_encode_args=self.tokenizer_encode_args,
            tokenizer_decode_args=self.tokenizer_decode_args,
            generation_args=generation_args,
            disable_batch_size_cache=self.disable_batch_size_cache,
            disable_generation_param_checks=self.disable_generation_param_checks,
            output_preproc=self.estimator.output_preproc,
            callbacks=self.estimator.callbacks_after_inference,
        )
        score = self.score_func(y_true=dataset_to_evaluate["_y"], y_pred=y_pred)
        return score, y_pred

    def get_value_at_quantile(self, input_list: List[str], quantile: float) -> int:
        """Get value at a specific quantile

        Args:
            input_list (List[str]): list of str on which to run the encoding of the `tokenizer`
            quantile (float): quantile on which to find the value on.

        Returns:
            int: rounded value at quantile
        """
        input_ids = self.tokenizer(
            input_list, max_length=None, truncation=False, padding=False
        )["input_ids"]
        batch_ids = list(map(len, input_ids))
        return round(np.quantile(batch_ids, q=quantile))



"""
Generation Related Utilties\n
**Note** : Generation Type Detection Scripts are updated as of transformers `v4.31.0`
"""

from typing import Dict, Union, List, Tuple

from llmsearch.utils.logging_utils import get_logger

logger = get_logger(__name__)

ops = ["equal", "lt", "gt", "ne"]

# Default generation parameters
gen_param_defaults = {
    "do_sample": False,
    "num_beams": 1,
    "num_beam_groups": 1,
    "penalty_alpha": None,
    "temperature": 1.0,
    "top_k": 50,
    "top_p": 1.0,
    "typical_p": 1.0,
    "epsilon_cutoff": 0.0,
    "eta_cutoff": 0.0,
    "diversity_penalty": 0.0,
    "repetition_penalty": 1.0,
    "encoder_repetition_penalty": 1.0,
    "length_penalty": 1.0,
    "no_repeat_ngram_size": 0,
    "bad_words_ids": None,
    "force_words_ids": None,
    "renormalize_logits": False,
    "constraints": None,
    "forced_bos_token_id": None,
    "forced_eos_token_id": None,
    "remove_invalid_values": None,
    "exponential_decay_length_penalty": None,
    "suppress_tokens": None,
    "begin_suppress_tokens": None,
    "forced_decoder_ids": None,
    "sequence_bias": None,
    "guidance_scale": None,
    "encoder_no_repeat_ngram_size": 0,
}


# Generation type mapping
gen_type_map = {
    "Contrastive Search Decoding": {
        "num_beams": {"equal": 1},
        "top_k": {"ne": None, "gt": 1},
        "do_sample": {"equal": False},
        "penalty_alpha": {"ne": None, "gt": 0},
    },
    "Greedy Decoding": {
        "num_beams": {"equal": 1},
        "num_beam_groups": {"equal": 1},
        "do_sample": {"equal": False},
    },
    "Sampling": {
        "num_beams": {"equal": 1},
        "num_beam_groups": {"equal": 1},
        "do_sample": {"equal": True},
    },
    "Beam Search": {
        "num_beams": {"gt": 1},
        "num_beam_groups": {"equal": 1},
        "do_sample": {"equal": False},
    },
    "Beam Search with sampling": {
        "num_beams": {"gt": 1},
        "num_beam_groups": {"equal": 1},
        "do_sample": {"equal": True},
    },
    "Group Beam Search": {
        "num_beams": {"gt": 1},
        "num_beam_groups": {"gt": 1},
        "diversity_penalty": {"gt": 0},
        "do_sample": {"equal": False},
    },
    "Mirostat Sampling": {
        "mirostat_mode": {"equal": 2},
        "do_sample": {"equal": True},
    },
    "Tail-Free Sampling": {
        "tfs": {"gte": 0, "lte": 1.0},
        "do_sample": {"equal": True},
    },
    "Top-A Sampling": {
        "top_a": {"gte": 0, "lte": 1.0},
        "do_sample": {"equal": True},
    },
}

# Generation parameters dependent on 'do_sample` Parameter
param_dependent_on_sampling = [
    "temperature",
    "top_k",
    "top_p",
    "typical_p",
    "eta_cutoff",
    "mirostat_mode",
    "tfs",
    "top_a",
]


def check_sample_parameter(gen_params: Dict, gen_type_params: Union[None, List]):
    """
    Check if the `do_sample` parameter is set to True when other parameters that are dependent on sampling are present.

    Args:
        gen_params (Dict): Dictionary containing generation parameters.
        gen_type_params (Dict): Params to exclude if already being used by a generation type (`top_k` in Contrastive Search Decoding)

    Raises:
        UserWarning: Warns if dependent generation parameters are present without 'do_sample' set to True.
    """
    if not gen_params.get("do_sample", gen_param_defaults["do_sample"]) and any(
        param in gen_params.keys() for param in param_dependent_on_sampling
    ):
        params = [
            param
            for param in gen_params.keys()
            if param in param_dependent_on_sampling
            if param not in gen_type_params
        ]
        if params:
            logger.warning(
                "Sampling dependent generation parameter/s present - %s, please set `do_sample` to True, for these params to work.",
                params,
            )


def identify_and_validate_gen_params(gen_params: Dict):
    """
    Identify and validate the generation type based on provided generation parameters.

    Args:
        gen_params (Dict): Dictionary containing generation parameters.

    Returns:
        str: Name of the identified generation type.
    """
    ret_val = "unable_to_detect"
    for gen_type_name, rules in gen_type_map.items():
        check = check_if_gen_param_rules_satisfy(gen_params=gen_params, rules=rules)
        if check:
            check_sample_parameter(
                gen_params=gen_params,
                gen_type_params=rules.keys(),
            )
            ret_val = gen_type_name
            break
    return ret_val


def check_if_gen_param_rules_satisfy(gen_params: Dict, rules: Dict):
    """
    Check if the provided generation parameters satisfy the specified rules.

    Args:
        gen_params (Dict): Dictionary containing generation parameters.
        rules (Dict): Dictionary containing rules for parameter validation.

    Returns:
        bool: True if all rules are satisfied, False otherwise.
    """
    checks = []
    for param, param_req in rules.items():
        check = check_if_param_req_satisfy(
            gen_params=gen_params, param=param, param_req=param_req
        )
        checks.append(check)
    return sum(checks) == len(checks)


def check_if_param_req_satisfy(gen_params: Dict, param: str, param_req: Dict):
    """
    Check if a specific generation parameter satisfies the specified requirement.

    Args:
        gen_params (Dict): Dictionary containing generation parameters.
        param (str): Name of the generation parameter to check.
        param_req (Dict): Dictionary specifying the requirement for the parameter.

    Returns:
        bool: True if the requirement is satisfied, False otherwise.
    """
    checks = []
    for op, value_to_check in param_req.items():
        check = False
        assert op in ops, f"Invalid operation - {op}, not in defined ops - {ops}"
        value = gen_params.get(param, gen_param_defaults[param])
        if op == "equal":
            check = value == value_to_check
        elif op == "ne":
            check = value != value_to_check
        elif op == "gt" and value is not None:
            check = value > value_to_check
        elif op == "lt" and value is not None:
            check = value < value_to_check
        checks.append(check)
    return bool(sum(checks))


def get_sample_hyp_space(seed: int, max_new_tokens: int) -> Tuple[List, List]:
    """Get 2 sample hyp spaces

    Args:
        seed (int): seed

    Returns:
        Tuple[List, List]: First Item is a larger hyp space which searches for individual generation types, Second Item of the Tuple are the top generation params as evaluated by [oobabooga using Vicuna-13B with instruct prompts](https://www.reddit.com/r/LocalLLaMA/comments/14adfw2/preset_arena_17205_comparisons_between_241/).
    """
    param_grid1 = [
        {
            "num_beams": [1],
            "top_k": list(range(4, 20)),
            "do_sample": [True],
            "penalty_alpha": [0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            "num_beams": [1],
            "num_beam_groups": [1],
            "do_sample": [False],
            "max_new_tokens": [max_new_tokens],
        },
        {
            "num_beams": [3],
            "num_beam_groups": [1],
            "do_sample": [False],
            "max_new_tokens": [max_new_tokens],
        },
        {
            "num_beams": list(range(2, 4)),
            "num_beam_groups": [1],
            "do_sample": [False],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
            "no_repeat_ngram_size": [0, 3, 4, 5],
        },
        # disabling as currently not supported
        # {
        #     "mirostat_mode": [2],
        #     "mirostat_eta": [0.1, 0.2, 0.3, 0.4, 0.5],
        #     "mirostat_tau": list(range(3, 8)),
        #     "do_sample": [True],
        #     "generation_seed": [seed],
        #     "max_new_tokens" : [max_new_tokens],
        # },
        {
            "tfs": [0.8, 0.85, 0.9, 0.95, 0.99],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            "top_a": [0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # exhaustive search
            "top_p": [0.5, 0.7, 0.9, 1.0],
            "temperature": [0.7, 0.8, 0.9, 1.0],
            "no_repeat_ngram_size": [0, 3, 4, 5],
            "repetition_penalty": [1.0, 1.1, 1.2],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
    ]
    # presets as found in https://github.com/oobabooga/oobabooga.github.io/blob/main/arena/results.md#presets-that-i-chose
    param_grid_2 = [
        {
            # random_preset_066
            "epsilon_cutoff": [1.49],
            "eta_cutoff": [10.42],
            "repetition_penalty": [1.17],
            "temperature": [1.31],
            "top_a": [0.52],
            "top_k": [49],
            "top_p": [0.14],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # random_preset_134
            "repetition_penalty": [1.01],
            "temperature": [0.87],
            "tfs": [0.68],
            "top_k": [85],
            "top_p": [0.99],
            "typical_p": [0.68],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # simple-1
            "repetition_penalty": [1.15],
            "temperature": [0.7],
            "top_k": [20],
            "top_p": [0.9],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # random_preset_035
            "repetition_penalty": [1.09],
            "temperature": [1.31],
            "top_k": [72],
            "top_p": [0.29],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # starchat
            "temperature": [0.2],
            "top_k": [50],
            "top_p": [0.95],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # random_preset_183
            "encoder_repetition_penalty": [1.07],
            "eta_cutoff": [10.78],
            "repetition_penalty": [1.21],
            "temperature": [1.01],
            "top_a": [0.75],
            "top_k": [91],
            "top_p": [0.21],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # tfs-with-top-a
            "repetition_penalty": [1.15],
            "temperature": [0.7],
            "tfs": [0.95],
            "top_a": [0.2],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        {
            # Special-Contrastive Search-3
            "do_sample": [False],
            "penalty_alpha": [0.3],
            "top_k": [4],
            "max_new_tokens": [max_new_tokens],
        },
        {
            "repetition_penalty": [1.02],
            "temperature": [1.68],
            "tfs": [0.97],
            "top_a": [0.42],
            "top_k": [77],
            "top_p": [0.17],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
        # disabling as currently not supported
        # {
        #     # mirostat
        #     "mirostat_mode": [2],
        #     "mirostat_eta": [0.1],
        #     "mirostat_tau": [5],
        #     "do_sample": [True],
        #     "generation_seed": [seed],
        #     "max_new_tokens" : [max_new_tokens],
        # },
        {
            # NovelAI-Ouroboros
            "repetition_penalty": [1.05],
            "temperature": [1.07],
            "top_k": [100],
            "do_sample": [True],
            "generation_seed": [seed],
            "max_new_tokens": [max_new_tokens],
        },
    ]
    return param_grid1, param_grid_2



"""
Common utilties
"""

import json
import yaml
import inspect
from pathlib import Path
from sklearn.base import BaseEstimator
from datasets import Dataset


def json_dump(ob: dict, file_path: Path):
    with open(file_path, "w", encoding="utf-8") as json_file:
        json.dump(ob, json_file, indent=4)


def json_load(file_path: Path):
    with open(file_path, "r", encoding="utf-8") as json_file:
        return json.load(json_file)


def yaml_load(file_path):
    """Load yaml file from file path

    Args:
        file_path (str): path to yaml file

    Returns:
        dict: loaded yaml file
    """
    with open(file_path, "r") as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def process_dataset_with_map(
    dataset, sample_preprocessor, tokenizer, input_cols, eval_cols
):
    """Processes the given dataset by mapping a processing function over each sample.

    Args:
        dataset (Dataset): A Hugging Face dataset to be processed.
        sample_preprocessor (function): A function to preprocess sample inputs.
        tokenizer (function): A tokenizer function to apply to input text.
        input_cols (list of str): Column names to be processed for input features.
        eval_cols (list of str): Column names for evaluation labels.

    Returns:
        Dataset: A new dataset with original data and additional processed keys `_X` and `_y`.
    """

    def process_sample(sample):
        # Copy original data to retain all keys
        sample_dict = dict(sample)
        processed_sample = sample_dict.copy()
        # Process inputs and store in '_X'
        processed_sample["_X"] = sample_preprocessor(
            tokenizer, **{col: sample[col] for col in input_cols + eval_cols}
        )
        # Extract evaluation columns and store in '_y'
        processed_sample["_y"] = {eval_col: sample[eval_col] for eval_col in eval_cols}
        return processed_sample

    # Apply the processing function to each sample in the dataset
    mapped_data = map(process_sample, dataset)
    # Convert the mapped data to a list (necessary for further processing)
    mapped_list = list(mapped_data)

    # Construct a new dataset from the list of processed samples
    # This uses dictionary comprehension to handle potentially complex nested structures in '_X' and '_y'
    return Dataset.from_dict(
        {key: [dic[key] for dic in mapped_list] for key in mapped_list[0]}
    )


def print_call_stack(n: int):
    """Prints call stack from first call to latest call

    Args:
        n (int): last n calls, last call would be the latest one
    """
    stack = inspect.stack()[1 : n + 1]
    for level, frame_info in enumerate(stack[::-1]):
        frame = frame_info.frame
        function_name = frame.f_code.co_name
        filename = frame.f_code.co_filename
        line_number = frame.f_lineno
        print(f"Level {level}: {function_name} | File: {filename}, Line: {line_number}")


def clone_monkey_patch(
    estimator: BaseEstimator, *, safe: bool = True  # pylint: disable=unused-argument
) -> BaseEstimator:
    """Deprecated Monkey Patch function to clone the Estimator while doing cross validation/hyperparameter search

    Usable in < 1.3 versions of scikit-learn versions

    - This functions returns the same estimator, as there are no parameters to specifically "fit"
    - This is done to avoid OOM errors for larger models, this does not affect the hyperparameter search in any way

    Args:
        estimator (BaseEstimator): estimator to clone
        safe (bool, optional): redundant parameter for now. Defaults to True.

    Returns:
        BaseEstimator: returns the same estimator
    """
    assert not hasattr(
        estimator, "model_generation_param_keys"
    ), f"Hyperparameters to tune already defined - {estimator.model_generation_param_keys}"
    return estimator



"""

<a href="https://freeimage.host/"><img src="https://iili.io/JpEgiIS.png" alt="JpEgiIS.png" border="0" /></a>

[![GitHub](https://img.shields.io/badge/github-%23121011.svg?style=for-the-badge&logo=github&logoColor=white)](https://github.com/Praful932/llmsearch)
[![Netlify Status](https://api.netlify.com/api/v1/badges/90e52247-da13-447b-b2eb-8f169a514877/deploy-status)](https://app.netlify.com/sites/llmsearch/deploys)
[![Hits](https://hits.seeyoufarm.com/api/count/incr/badge.svg?url=https%3A%2F%2Fgithub.com%2FPraful932%2Fllmsearch&count_bg=%2379C83D&title_bg=%23555555&icon=&icon_color=%23E7E7E7&title=hits&edge_flat=false)](https://hits.seeyoufarm.com)


Conduct hyperparameter search over generation parameters of large language models (LLMs). This tool is designed for ML practitioners looking to optimize their sampling strategies to improve model performance. Simply provide a model, dataset, and performance metric, llmsearch handles the rest.

### QuickStart
-  llama-3-8b Example [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/Praful932/llmsearch/blob/main/examples/llmsearch_quickstart.ipynb) - A quickstart notebook which shows the basic functionality of llmsearch. This notebook will help you understand how to quickly set up and run hyperparameter searches.

### End-to-End Model Examples
1. [GSM8K Example](https://github.com/Praful932/llmsearch/blob/main/examples/gsm8k_example.ipynb) - Shows a `GridSearchCV` ran on the [GSM8K](https://huggingface.co/datasets/gsm8k) Dataset using the `TheBloke/CapybaraHermes-2.5-Mistral-7B-AWQ` model.
2. [Samsum Example](https://github.com/Praful932/llmsearch/blob/main/examples/samsum_example.ipynb) - Shows a `GridSearchCV` ran on the [samsum](https://huggingface.co/datasets/samsum) Dataset
using a finetuned(on the same dataset) version of `cognitivecomputations/dolphin-2.2.1-mistral-7b`.

Refer [README](https://github.com/Praful932/llmsearch) further for more details on how to use `llmsearch`.

# API Reference Documentation
"""
from llmsearch.patches.transformers_monkey_patch import hijack_samplers

print("Monkey Patching .generate function of `transformers` library")
hijack_samplers()

__version__ = ""


"""Stopping criteria utilities for generation."""

from typing import List

import torch
from transformers import StoppingCriteria


class MultiTokenStoppingCriteria(StoppingCriteria):
    """Criteria to stop on the specified multi-token sequence.
    A modified verion of Stopping Criteria forked and modified from [lm-evaluation-harness](https://github.com/EleutherAI/lm-evaluation-harness/blob/27924d77953491f66a038a09892807065e469358/lm_eval/models/utils.py#L208)
    That maintains a state for each batch of inputs which helps in knowing where to look.
    """

    def __init__(
        self,
        sequence_ids: List[int],
    ) -> None:
        self.sequence_ids = torch.tensor(
            sequence_ids, dtype=torch.int32, device="cuda:0"
        )
        # we look back for 2 more tokens than it takes to encode our stop sequence
        # because tokenizers suck, and a model might generate `['\n', '\n']` but our `sequence` is `['\n\n']`
        # and we don't want to mistakenly not stop a generation because our
        # (string) stop sequence was output in a different tokenization
        # NOTE: there is a minor danger that this will end up looking back 2 tokens into the past, into the inputs to the model,
        # and stopping generation immediately as a result. With only 2 extra tokens of lookback, this risk is minimized
        # Additionally, in lookback_ids_batch we should prevent ever looking back into the inputs as described.
        self.sequence_id_len = self.sequence_ids.shape[0] + 2
        self.state_initialized = False
        self.prompt_length = None
        self.state_initialized = False

        self.batch_size, self.done_tracker = None, []

    def set_state(self, batch_size, prompt_length):
        """Set state before starting generation for a new batch"""
        self.batch_size = batch_size
        self.prompt_length = prompt_length
        self.done_tracker = [False] * batch_size
        self.state_initialized = True

    def reset(self):
        """Reset state before starting generation for a new batch"""
        self.batch_size, self.done_tracker = None, []
        self.prompt_length = None
        self.state_initialized = False

    def __call__(self, input_ids, scores, **kwargs) -> bool:
        """
        This is called after a new token is generated
        """
        ret_val = False

        if not self.state_initialized:
            # Every batch should set this state
            self.set_state(input_ids.shape[0], input_ids.shape[1] - 1)
        if input_ids.shape[0] != self.batch_size:
            self.reset()
            # last inference failed, set the state again, prompt length is the same
            # TODO : self.done_tracker can be cached here but not sure if it is worth it
            self.set_state(input_ids.shape[0], self.prompt_length)

        # IDs of all the tokens except the prompt
        lookback_ids_batch = input_ids[:, self.prompt_length :]

        # total_tokens_till_now = lookback_ids_batch.shape[1]
        # look back for 2 more tokens than it takes to encode our stop sequence
        lookback_ids_batch = lookback_ids_batch[:, -self.sequence_id_len :]

        # no elements yet to look back
        if lookback_ids_batch.nelement() == 0:
            return False

        # for each item in the batch
        for i, done in enumerate(self.done_tracker):
            if not done:
                # look back as much as the length of the stop token sequence
                self.done_tracker[i] = (
                    self.sequence_ids
                    == lookback_ids_batch[i][-(self.sequence_ids.shape[0]) :]
                )

        ret_val = False not in self.done_tracker
        if ret_val:
            # useful to know when the generation stops
            pass
            # self.reset()
        return ret_val
